/*----------------------------------------------*/
/*						*/
/*		ADDTOCASEBASE.C			*/
/*						*/
/* This file contains procedures for adding the	*/
/* repaired case to the casebase. It asks for a	*/
/* new name for the case and a value for the	*/
/* result. The case stored uses the problem	*/
/* fields from the usercase and the solution	*/
/* fields from the repair case.			*/
/*						*/
/* Modified for version 36.			*/
/*	The result value has been abolished	*/
/*----------------------------------------------*/

/* prompt user to update casebase */

void updatecasebase()
{
char reply[10];

 printf("\n\n\n");
 do
   {
    printf("Do you want to add the new case to the casebase? (yes,no): ");
    getstring(reply);
   }
 while (strcmp(reply,"yes") != 0 && strcmp(reply, "no") != 0);

 if (strcmp(reply,"yes") == 0)
   {
    addrepairedcase();
   }
}


/* add the repaired usercase to the casebase */

void addrepairedcase()
{
struct casenode *newcase;
buftype casename;

 newcase = copycase(repcase);	/* creates a new case */
 printf("\n\n\n");
 
 while (TRUE)   /* loop forever */
   {
    printf("Enter a name for the case: ");
    getstring(casename);
    if (casenameexists(casename) == TRUE)
      {
       printf("Casename %s already exists\n",casename);
      }
    else
      {
       strcpy(newcase->name,casename);
       break;
      }
   }
      
 printf("\n\n\n");

 addtocasebase(newcase);	/* this does NOT make a copy */
}


/* return TRUE if a case with the same name already exists */

int casenameexists(casename)
buftype *casename;
{
struct casenode *pointer;

 pointer = caseroot;

 while (pointer != NULL)
   {
    if (strcmp(pointer->name,casename) == 0)	/* found matching name */
      {
       return TRUE;
      }
    pointer = pointer->caselink;
   }

 return FALSE;
}


/* create a new case containing the given name and field lists */

struct casenode *assemblecase(name,problemfields,solnfields)
buftype *name;
struct fieldnode *problemfields, *solnfields;
{
struct casenode *pointer;

 pointer = createcasenode(name);
 pointer->fieldhead = copyfieldlist(problemfields);
 pointer->solnhead = copyfieldlist(solnfields);
 pointer->fielddefhead = NULL;	/* local field defs not copied */
 pointer->rephead = NULL;	/* local repair rules not copied */
 return pointer;
}


/* link the case to the head of the casebase and add indexes */

void addtocasebase(caseptr)
struct casenode *caseptr;
{
 caseptr->caselink = caseroot;
 caseroot = caseptr;	/* add to head of list */
 modifyindexes();	/* add indexes for case at head of list */
 appendtofile();	/* append case to case file */
}


/* add indexes for case at head of case list */

void modifyindexes()
{
struct indexnamenode *pointer;

 pointer = indexnameroot;

 while (pointer != NULL)
   {
    modifyindex(pointer->fieldname,pointer->indexhead);
    pointer = pointer->indexnamelink;
   }
}


/* modify index for case at head of list */

void modifyindex(fieldname, indexptr)
buftype *fieldname;
struct indexnode *indexptr;
{
		/* if field value matches this index node */
 if (strcmp(indexptr->evalue, getfieldvalue(caseroot,fieldname)) == 0)
   {
    addnewcaseptr(indexptr);	/* put case pointer in */
   }
 else
   {
    if (indexptr->indexlink == NULL)
      {
       addnewindex(fieldname,indexptr);	/* add new index to tail of list */
      }
    else
      {
       modifyindex(fieldname, indexptr->indexlink);	/* look at tail */
      }
   }
}


/* add a new case pointer at head of list, pointing to new case */

void addnewcaseptr(indexptr)
struct indexnode *indexptr;
{
struct caseptrnode *pointer;

 pointer = createcaseptrnode(caseroot);
 pointer->caseptrlink = indexptr->caseptrhead;
 indexptr->caseptrhead = pointer;
}


/* add an index node to the tail and create a case pointer to the new case */

void addnewindex(fieldname,indexptr)
buftype *fieldname;
struct indexnode *indexptr;
{
struct indexnode *pointer;

 pointer = createindexnode(getfieldvalue(caseroot,fieldname));
 addnewcaseptr(pointer);
 indexptr->indexlink = pointer;
}


/* append new case to case file */
void appendtofile()
{
int temp;

 temp = listexpansion;
 listexpansion = FALSE;
 fp = fopen(casefile,"a");	/* open the case file for appending */
 displaycase(caseroot);		/* writes the case out */
 fclose(fp);			/* close the file tidy like */
 listexpansion = temp;
 fp = stdout;
}
