#define TRUE 1
#define FALSE 0

#define STRTYPE 1
#define NUMTYPE 2
#define ENUMTYPE 3
#define LISTTYPE 4
#define LISTOPTYPE 5
#define EXTRACTOPTYPE 6
#define STRBUFSIZE 51	/* maximum length of string/symbol is STRBUFSIZE - 1 */
#define SIMWEIGHT 0.75
#define MAXITER	1000	/* maximum no of iterations for outer repair loop */

/* definitions for arithnode contents type */
#define NUMVALUE 0
#define FIELDNAME 1
#define PLUS 2
#define MINUS 3
#define MULTIPLY 4
#define DIVIDE 5
#define UMINUS 6
#define POWER 7

/* definitions for combtreenode operation type */
#define ITEM 0
#define OR_OP 1
#define AND_OP 2
#define NOT_OP 3

/* definitions for listopnode contentstype */
#define HEAD 2
#define TAIL 3

/* the buftype is used to store strings and enumeration values as 
   strings. When the fields are defined, the types are stored in
   a linked list of typenodes. If a type is an enumeration, then
   the typenode will point to the head of a linked list of
   enumnodes, which are strings containing the enumeration types.
   The case instances are stored as a linked list of casenodes.
   Each casenode holds the name of the case, the result for the case
   and a pointer to a linked list of fieldnodes which describe the
   contents of the fields of a particular case. The fieldnode
   holds the name and type of each field and the contents of the
   field.

   The STACKTYPE is the type of the YACC value-stack. It is a union
   of various different types needed to construct the linked lists

   This version has been modified to store the solution fields as
   a separate list, for each case instance.

   This version has been modified for version 3 which includes the
   special goal field

   This version has been modified for version 4 which does checking
   for doubly defined fields in the field definition and multiple
   uses of a field in each case instance. In a case instance a field
   may be given a value once in the problem section and once in the
   solution section.

   This version has been modified for version 5 which implements
   lists. The type of a list member may be of any type including another
   list.

   This version has been modified for version 6 which checks for
   doubly defined case names

   This version has been modified for version 7 which checks for
   non - unique enumeration symbols

   This version has been modified for version 8 which does simple
   case matching

   This version has been modified for version 9 which has a weight value
   added to the typenode structure.

   This version has been modified for version 10 which has modification
   rules added

   This version has been modified for version 11 which incorporates
   repair rules

   This version has been modified for version 12 which incorporates
   arithmetical expressions in repair rules

   This version has been modified for version 13 which incorporates
   the NEW VALUE concept in repair rules

   This version has been modified for version 14 in which the repair rule
   code has been re-written for the more general format rules

   This version has been modified for version 15 which has the GOAL
   field removed.

   This version has been modified for version 16 which has the ABSTRACTION
   mechanism

   This version has been modified for version 17 which has indexing added.

   This version has been modified for version 18 which has had a general
   tidy up of the code, and does a check for doubly defined abstraction
   symbols. It also now displays the casebase in the same format as the
   input file.

   This version has been modified for version 19 which introduces a menu.

   This version has been modified for version 20 which allows the repaired
   usercase to be added to the casebase.

   This version has been modified for version 21 which returns to the
   main menu after a match/repair/store cyle has been completed

   This version has been modified for version 22 which does list expansion

   This version has been modified for version 23 which has the NEW VALUE
   mechanism removed.

   This version has been modified for version 24 which has the introduction
   block added.

   This version has been modified for version 25 which has changes made to
   the way enumeration values are entered.

   This version has been modified for version 26 which has the REPAIR action
   added to the repair rules and allows modification of problem fields.

   This version has been modified for version 27 which changes the
   weight returned when comparing numerical fields.

   This version has been modified for version 28 which implements the
   weight rules;

   This version has been modified for version 29 which implements the
   comparison operation in combnodes.

   This version has been modified for version 30 which implements text
   matching.

   This version has been modified for version 32 which implements local
   repair rules.

   This version has been modified for version 34 which abolishes the
   solution definition.

   This version has been modified for version 36 which abolishes the
   result value.

   This version has been modified for version 38 which implements
   the basic list operations hd and tl

   This version has been modified for version 39 which implements
   the reselect action.

   This version has been modified for version 40 which implements
   local field definitions.

   This version has been modified for version 42 which implements
   list matching for non-identical lists

   This version has been modified for version 43 which implements
   the extract operation (repair action).

   This version has been modified for version 44 which alters the
   extract operation to delete the symbol from the list after extraction.

   This version has been modified for version 45 which implements
   the prompt feature.

   This version has been modified for version 46 which implements
   proper logical functions and,or,not and ().

   This version has been modified for version 48 which modifies
   matching of lists. enterfield() and enterfields() are now void type

   This version has been modified for version 49 which replaces
   weight rules with preprocess rules.

*/


typedef char buftype[STRBUFSIZE];


struct listnode	/* holds values of list items */
  {
   buftype *strcontents;
   float numcontents;
   struct listnode *listitem;
   int listtype;
   struct listnode *listlink;
  };

struct listopnode	/* defines a list operation node */
  {
   int optype;		/* defines the operation type */
   buftype *fieldname;	/* pointer to fieldname string if optype = FIELDNAME*/
   struct listopnode *listoplink;	/* pointer to nested list op */
  };

struct fieldnode	/* defines case field */
  {
   buftype *strcontents;
   float numcontents;
   struct listnode *listcontents;
   buftype fieldname;
   int fieldtype;
   struct fieldnode *fieldlink;
  };

struct enumnode		/* holds enumeration value as a string */
  {
   buftype evalue;
   struct enumnode *enumlink;
  };

struct typenode		/* defines a field type */
  {
   buftype fieldname;
   int fieldtype;
   struct enumnode *enumhead;
   float weight;
   float oldweight;
   struct listnode *prompt;	/* defines a list to hold a prompt */
   struct typenode *typelink;
  };

struct modnode		/* defines a modification rule definition */
  {
   buftype fieldname;
   int fieldtype;	/* enumeration list or number range */
   float start,end;	/* define number range */
   struct enumnode *enumhead;	/* pointer to list of enumeration symbols */
   buftype abstraction;	/* enumeration abstraction symbol */
   struct modnode *modlink;	/* pointer to next modnode */
  };

struct arithnode        /* node for arithmetical expression tree */
  {
   int contentstype;    /* indicates the type of the contents of this node */
   float numcontents;  /* the numerical value */ 
   buftype fieldname;   /* name of field containing number value */ 
   struct arithnode *left,*right;       /* pointers to subtrees */
  };

struct replistnode	/* defines repair action list node */
  {
   buftype fieldname;
   int fieldtype;
   int undefined;	/* boolean, TRUE if field becomes undefined */
   int restart;		/* boolean, TRUE if REPAIR action */
   int print;		/* boolean, TRUE if print statement */
   int reselect;	/* boolean, TRUE if RESELECT action */
   int listop;		/* boolean, TRUE if list operation */
   int extract;		/* boolean, TRUE if EXTRACT action */
   int weightchange;	/* boolean, TRUE if change weight op */
   float weight;	/* new weight value */
   struct arithnode *exptree;	/* pointer to expression tree */
   buftype *abstraction;	/* used for EXTRACT operation */
   buftype *strvalue;	/* string/enumeration to print or for field to take */
   struct listnode *tolist;	/* new list value for field to take */
   struct listopnode *listophead;	/* pointer to list operation */
   struct replistnode *replistlink;	/* pointer to next replistnode */
  };


struct combnode		/* defines node to hold combination of field values */
  {
   buftype fieldname;
   int fieldtype;	/* if NUMTYPE then denotes comparison operation */
   int undefined;	/* boolean, TRUE if field becomes undefined */
   int listop;		/* true if list operation */
   struct arithnode *left;	/* left side of comparison */
   struct arithnode *right;	/* right side of comparison */
   int operator;		/* comparison operator */
   buftype strcontents;	/* string/enumeration value */
   struct listnode *listvalue;	/* pointer to list value */
   struct listopnode *listophead;	/* pointer to list operation */
  };

struct combtreenode	/* defines the combination tree node */
  {
   int optype;		/* defines the logical operation */
   struct combtreenode *left,*right;	/* pointer to child nodes */
  };

struct repnode		/* defines a repair rule */
  {
   buftype rulename;
   struct combtreenode *combtreeroot;/* pointer to root of combination tree */
   struct replistnode *replisthead;/* points to head of repair action list */
   struct repnode *replink;	/* pointer to next repair rule */
  };

struct casenode         /* defines a case */
  {
   buftype name;
   struct typenode *fielddefhead;	/* local field definitions */
   struct fieldnode *fieldhead;
   struct fieldnode *solnhead;	/* point to head of solution fields */
   struct repnode *rephead;	/* pointer to local repair rule list */
   struct casenode *caselink;
  };

struct rulenamenode	/* holds a rulename */
  {
   buftype rulename;
   struct rulenamenode *rulenamelink;
  };

struct caseptrnode	/* holds a pointer to a case */
  {
   struct casenode *caseind;	/* pointer to the case */
   struct caseptrnode *caseptrlink;	/* pointer to next caseptr node */
  };

struct indexnode	/* holds list of case pointers for a value */
  {
   buftype evalue;	/* the enumeration value */
   struct caseptrnode *caseptrhead;	/* head of caseptr list */
   struct indexnode *indexlink;		/* pointer to next index node */
  };

struct indexnamenode	/* holds a name of an index */
  {
   buftype fieldname;	/* name of field to create an index on */
   struct indexnode *indexhead;	/* head of index node list */
   struct indexnamenode *indexnamelink;	/* points to next node */
  };

 struct indexspecnode	/* holds an index field & value specification */
   {
    buftype fieldname;
    buftype value;
    struct indexspecnode *indexspeclink;/* pointer to next indexspecnode */
   };

 struct intronode	/* holds introduction text */
   {
    buftype *text;
    struct intronode *introlink;
   };

/* Global variables */

int listexpansion;	/* set to FALSE by default */
jmp_buf environment;	/* used to restore stack after longjmp */
int START;		/* TRUE = at start of case file */
FILE *fp;	/* pointer to case file */
struct intronode *introroot;	/* linked list of introduction text */
struct typenode *probdefnroot;	/* list of problem field definitions */
struct casenode *caseroot;	/* list of cases */
struct caseptrnode *caseptrroot;	/* list of pointers to cases */
struct casenode *usercase;	/* the users case */
struct casenode *foundcase;	/* the case found by the casematching */
struct casenode *repcase;	/* the repaired case */
struct modnode *modroot;	/* the list of modification rules */
struct repnode *preprocessroot;	/* the list of pre-process repairs */
struct repnode *reproot;	/* the list of repair rules */
struct indexnamenode *indexnameroot;	/* the list of index names */
char ch;			/* character used by both lexical analysers */
long lines;		/* incremented each time newline is met (lex.c) */
buftype lexeme;		/* lexeme returned by gettok */
int token;		/* token returned by gettok */
char casefile[30];	/* name of the file containing the case base */
int restartrepair;	/* TRUE = restart the repair rules */
int reselectcase;	/* TRUE = reselect case and repair */

/* Note the values  for the following two variables are created by
   similarrange and used by compnumfields in casematching.c. They
   should NOT be used elsewhere */

float upperbound;	/* upper bound of current similar range */
float lowerbound;	/* lower bound of current similar range */

/* function declarations for those which don't return an integer */
char *createstringnode();
char *enumchoice();
char *dolistop();
buftype *createbuftypenode();
buftype *getfieldvalue();
struct intronode *createintronode();
struct enumnode *searchtypelist();
struct enumnode *searchtypelistok();
struct fieldnode *createnewfieldnode();
struct fieldnode *createstrfield();
struct fieldnode *createnumfield();
struct fieldnode *createenumfield();
struct fieldnode *createlistfield();
struct fieldnode *enterstrvalue();
struct fieldnode *enternumvalue();
struct fieldnode *enterenum();
struct fieldnode *enterenummenu();
struct fieldnode *enterlistvalue();
struct fieldnode *entertext();
struct fieldnode *searchfield();
struct fieldnode *copyfieldlist();
struct fieldnode *copyfield();
struct repnode *createrepnode();
struct combnode *createcombenumnode();
struct combnode *createcombcompnode();
struct combnode *createcombstringnode();
struct combnode *createcomblistnode();
struct combnode *createcomblistopenumnode();
struct combnode *createcomblistopexpnode();
struct combnode *createcomblistopstringnode();
struct combnode *createcomblistoplistnode();
struct combnode *createcombundefnode();
struct combtreenode *createcombtreenode();
struct replistnode *createreplistenumnode();
struct replistnode *createreplistexpnode();
struct replistnode *createreplistlistnode();
struct replistnode *createreplistlistopenumnode();
struct replistnode *createreplistlistopexpnode();
struct replistnode *createreplistlistopstringnode();
struct replistnode *createreplistlistoplistnode();
struct replistnode *createrepliststringnode();
struct replistnode *createreplistundefnode();
struct replistnode *createreplistrepnode();
struct replistnode *createreplistprintnode();
struct replistnode *createreplistresnode();
struct replistnode *createreplistextractnode();
struct replistnode *createreplistweightnode();
struct listopnode *createlistopnode();
struct listopnode *createlistopfieldnode();

float findmatchfields();
float domatch();
float compstrfields();
float compenumfields();
float compnumfields();
float complistfields();
float matchkeys();
float matchlist();
float matchsublist();
float findweight();
float evaluate();
float fieldvalue();
struct indexspecnode *createindexspecnode();
struct indexspecnode *specifyindexes();
struct indexspecnode *makeindexspec();
struct indexnamenode *createindexnamenode();
struct indexnamenode *findindexname();
struct indexnode *createindexnode();
struct indexnode *findindex();
struct caseptrnode *createcaseptrnode();
struct caseptrnode *makecaseptrlist();
struct caseptrnode *buildmatchlist();
struct caseptrnode *searchabstract();
struct caseptrnode *copycaseptrlist();
struct caseptrnode *matchenumlist();
struct arithnode *createplusnode();
struct arithnode *createminusnode();
struct arithnode *createmultiplynode();
struct arithnode *createdividenode();
struct arithnode *createpowernode();
struct arithnode *createunaryminusnode();
struct arithnode *createnumbernode();
struct arithnode *createfieldvaluenode();
struct casenode *createcasenode();
struct casenode *matchcase();
struct casenode *findcase();
struct casenode *copycase();
struct casenode *browsecaselist();
struct casenode *assemblecase();
struct casenode *assemblerepcase();
struct enumnode *createenumnode();
struct typenode *createtypenode();
struct typenode *findfield();
struct listnode *createlistenum();
struct listnode *createlistnum();
struct listnode *createliststring();
struct listnode *createlistlist();
struct listnode *parsethelist();
struct listnode *parselist();
struct listnode *parsesequence();
struct listnode *createtextlist();
struct listnode *copylist();
struct modnode *createmodenumnode();
struct modnode *createmodrangenode();
void match();
void gettok();
void searchenodes();
void doublecheckcases();
void doublecheckenums();
void checkindexblock();	/* check the index block */
void checkmodrules();	/* check modification rules */
void checkabstraction();	/* check for doubly defined abstraction symbol */
void checkreselect();		/* check for reselect in pre-processing rules */
void checklocalreprules();	/* check local repair rules */
void checkreprules();	/* check repair rules */
void checkrepname();	/* check for doubly used repair rule names */
void checkcombnode();	/* check a combination item */
void checkcombtree();	/* check a combination tree */
void checkfieldexists();
void doublecheckreplist();
void doublecheckfields();
void multiplecheckfields();
void checkmultcase();
void checkfieldlist();
void checklocalfielddefs();
void typecheckfields();
void checkcase();
void checkfield();
void checkfieldtype();	/* check that named field is defined and has the specified type */
void localchecktype();	/* check type in both local and global definitions */
void checktype();
void checkenum();	/* check that enumeration value exists for specified fieldname */
void localcheckenum();	/* check enumeration value in local & global field definitions */
void extractenum();
void skipwhitespace();
void skipcomments();
void displaycase();
void displayfield();
void displaylist();
void displaylistitems();
void displayfieldexpand();
void casematching();
void entercase();
void getstring();
void rulerepair();
void repairfields();
void deletenode();
void deletenodeinfo();
void deletelist();
void firerule();
void firereplist();
void changeweight();
void changetostring();
void changetoenum();
void changetonum();
void changetolist();
void changetonumvalue();
void addrulename();
void deleterulenamelist();
void buildindexes();
void makeindex();
void addindexnode();
void indexsearch();
void performsearch();
void performfirstpass();
void performnextpass();
void intersection();
void getunion();
void deletecaseptrnode();
void addcaseptrtolist();
void deleteindexspec();
void initialise();
void menu();
void displaymenu();
void enterchoice();
void selectchoice();
void addrepairedcase();
void addtocasebase();
void modifyindexes();
void modifyindex();
void addnewcaseptr();
void addnewindex();
void updatecasebase();
void appendtofile();
void displayblurb();
void displayintro();
void cleanup();
void deletecaseptrlist();
void deletecase();
void deletefieldlist();
void deletefield();
void resetweights();
void displayenummenu();
void copyword();
void juxtaposed();
void doreselectcase();
void removecaseptr();
void deletelistitem();
void expandlist();
void enterfield();
void enterfields();
void preprocess();
